/* -*- tab-width: 4 -*- */

/* Ports and I/O primitives.
 */
#include "elkkernel.h"

#include <errno.h>
#ifdef ONUNIX
#  include <stdio.h>
#  include <pwd.h>
#endif
#if defined(MACOS) && !defined(MACOSX)
#  include <unix.h>
#else
#  include <sys/types.h>
#  include <sys/stat.h>
#endif

#ifndef WIN32
extern int errno;
#endif

ELKObject Get_File_Name (ELKObject name);
CHAR *Internal_Tilde_Expand (CHAR *s, CHAR **dirp);
ELKObject General_File_Operation (ELKObject s, int op);
ELKObject Open_File (CHAR *name, int flags, int err);
ELKObject General_Close_Port (ELKObject port);
ELKObject General_Call_With (ELKObject name, int flags, ELKObject proc);
ELKObject Curr_Input_Port, Curr_Output_Port;
ELKObject Standard_Input_Port, Standard_Output_Port;

void Init_Io (void) {
	Standard_Input_Port = Make_Port (P_INPUT, stdin, Make_String (x_("stdin"), 5));
	Standard_Output_Port = Make_Port (0, stdout, Make_String (x_("stdout"), 6));
	Curr_Input_Port = Standard_Input_Port;
	Curr_Output_Port = Standard_Output_Port;
	Global_GC_Link (Standard_Input_Port);
	Global_GC_Link (Standard_Output_Port);
	Global_GC_Link (Curr_Input_Port);
	Global_GC_Link (Curr_Output_Port);
}

void Reset_IO (int destructive) {
	Discard_Input (Curr_Input_Port);
	if (destructive)
		Discard_Output (Curr_Output_Port);
	else
		Flush_Output (Curr_Output_Port);
	Curr_Input_Port = Standard_Input_Port;
	Curr_Output_Port = Standard_Output_Port;
}

ELKObject Make_Port (int flags, FILE *f, ELKObject name) {
	ELKObject port;
	GC_Node;

	GC_Link (name);
	port = Alloc_Object (sizeof (struct S_Port), T_Port, 0);
	PORT(port)->flags = flags|P_OPEN;
	PORT(port)->file = f;
	PORT(port)->name = name;
	PORT(port)->ptr = 0;
	PORT(port)->lno = 1;
	PORT(port)->closefun = fclose;
	GC_Unlink;
	return port;
}

ELKObject P_Port_File_Name (ELKObject p) {
	Check_Type (p, T_Port);
	return (PORT(p)->flags & P_STRING) ? Falseob : PORT(p)->name;
}

ELKObject P_Port_Line_Number (ELKObject p) {
	Check_Type (p, T_Port);
	return Make_Unsigned (PORT(p)->lno);
}

ELKObject P_Eof_Objectp (ELKObject x) {
	return TYPE(x) == T_End_Of_File ? Trueob : Falseob;
}

ELKObject P_Current_Input_Port (void) { return Curr_Input_Port; }

ELKObject P_Current_Output_Port (void) { return Curr_Output_Port; }

ELKObject P_Input_Portp (ELKObject x) {
	return TYPE(x) == T_Port && IS_INPUT(x) ? Trueob : Falseob;
}

ELKObject P_Output_Portp (ELKObject x) {
	return TYPE(x) == T_Port && IS_OUTPUT(x) ? Trueob : Falseob;
}

int Path_Max (void) {
#if defined(WIN32) || defined(MACOS)		  /* POSIX */
	return 256;
#else
#ifdef PATH_MAX		  /* POSIX */
	return PATH_MAX;
#else
#ifdef MAXPATHLEN		/* 4.3 BSD */
	return MAXPATHLEN;
#else
#ifdef PATHCONF_PATH_MAX
	static r;
	if (r == 0) {
		if ((r = pathconf (x_("/"), _PC_PATH_MAX)) == -1)
			r = 1024;
		r++;
	}
	return r;
#else
	return 1024;
#endif
#endif
#endif
#endif
}

ELKObject Get_File_Name (ELKObject name) {
	register int len;

	if (TYPE(name) == T_Symbol)
		name = SYMBOL(name)->name;
	else if (TYPE(name) != T_String)
		Wrong_Type_Combination (name, x_("string or symbol"));
	if ((len = STRING(name)->size) > Path_Max () || len == 0)
		Primitive_Error (x_("invalid file name"));
	return name;
}

CHAR *Internal_Tilde_Expand (CHAR *s, CHAR **dirp) {
#if defined(WIN32) || defined(MACOS)
	return(0);
#else
	register CHAR *p;
	struct passwd *pw;

	if (*s++ != '~')
		return 0;
	for (p = s; *p && *p != '/'; p++)
		;
	if (*p == '/') *p++ = 0;
	if (*s == '\0') {
		if ((*dirp = egetenv (x_("HOME"))) == 0)
			*dirp = x_("");
	} else {
		pw = getpwnam(s);
		if (pw == 0)
			Primitive_Error (x_("unknown user: ~a"), Make_String (s, estrlen (s)));
		*dirp = pw->pw_dir;
	}
	return p;
#endif
}

ELKObject General_File_Operation (ELKObject s, register int op) {
	register CHAR *r;
	ELKObject ret, fn;
	int leng;
	Alloca_Begin;

	fn = Get_File_Name (s);
	Get_Strsym_Stack (fn, r);
	switch (op) {
	case 0: {
		CHAR *p, *dir;
		if ((p = Internal_Tilde_Expand (r, &dir)) == 0) {
			Alloca_End;
			return s;
		}
		leng = estrlen (dir) + 1 + estrlen (p) + 1;
		Alloca (r, CHAR*, leng);
		esnprintf (r, leng, x_("%s/%s"), dir, p);
		ret = Make_String (r, estrlen (r));
		Alloca_End;
		return ret;
	}
	case 1: {
		struct stat st;
		/* Doesn't make much sense to check for errno != ENOENT here:
		 */
		ret = stat (r, &st) == 0 ? Trueob : Falseob;
		Alloca_End;
		return ret;
	}}
	/*NOTREACHED*/
	return(Null);
}

ELKObject P_Tilde_Expand (ELKObject s) {
	return General_File_Operation (s, 0);
}

ELKObject P_File_Existsp (ELKObject s) {
	return General_File_Operation (s, 1);
}

void Close_All_Files (void) {
	Terminate_Type (T_Port);
}

ELKObject Terminate_File (ELKObject port) {
	(void)(PORT(port)->closefun) (PORT(port)->file);
	PORT(port)->flags &= ~P_OPEN;
	return Void;
}

ELKObject Open_File (CHAR *name, int flags, int err) {
	register FILE *f;
	CHAR *dir, *p;
	int leng;
	ELKObject fn, port;
	struct stat st;
	Alloca_Begin;

	if ((p = Internal_Tilde_Expand (name, &dir)) != 0)
	{
		leng = estrlen (dir) + 1 + estrlen (p) + 1;
		Alloca (name, CHAR*, leng);
		esnprintf (name, leng, x_("%s/%s"), dir, p);
	}
	if (!err && stat (name, &st) == -1
#ifndef MACOS
		&& (errno == ENOENT || errno == ENOTDIR)
#endif
		) {
		Alloca_End;
		return Null;
	}
	switch (flags & (P_INPUT|P_BIDIR)) {
	case 0:			   p = x_("w");  break;
	case P_INPUT:		 p = x_("r");  break;
	default:			  p = x_("r+"); break;
	}
	fn = Make_String (name, estrlen (name));
	Disable_Interrupts;
	if ((f = efopen (name, p)) == NULL) {
		Saved_Errno = errno;  /* errno valid here? */
		Primitive_Error (x_("~s: ~E"), fn);
	}
	port = Make_Port (flags, f, fn);
	Register_Object (port, (GENERIC)0, Terminate_File, 0);
	Enable_Interrupts;
	Alloca_End;
	return port;
}

ELKObject General_Open_File (ELKObject name, int flags, ELKObject path) {
	ELKObject port, pref;
	CHAR *buf = 0;
	register CHAR *fn;
	register int plen, len, blen = 0, gotpath = 0;
	Alloca_Begin;

	name = Get_File_Name (name);
	len = STRING(name)->size;
	fn = STRING(name)->data;
#ifdef MACOS
	if (fn[0] != '/' && fn[1] != ':' && fn[0] != '~')
#endif
#ifdef WIN32
	if (fn[0] != '/' && fn[1] != ':' && fn[0] != '~')
#endif
#ifdef ONUNIX
	if (fn[0] != '/' && fn[0] != '~')
#endif
	{
		for ( ; TYPE(path) == T_Pair; path = Cdr (path)) {
			pref = Car (path);
			if (TYPE(pref) == T_Symbol)
				pref = SYMBOL(pref)->name;
			if (TYPE(pref) != T_String)
				continue;
			gotpath = 1;
			if ((plen = STRING(pref)->size) > Path_Max () || plen == 0)
				continue;
			if (len + plen + 2 > blen) {
				blen = len + plen + 2;
				Alloca (buf, CHAR*, blen);
			}
			bcopy (STRING(pref)->data, buf, plen);
#ifdef MACOS
			if (buf[plen-1] != ':') buf[plen++] = ':';
#else
			if (buf[plen-1] != '/') buf[plen++] = '/';
#endif
			bcopy (fn, buf+plen, len);
			buf[len+plen] = '\0';
#ifdef MACOS
			if (buf[0] == '.' && buf[1] == ':') buf += 2;
#endif
			port = Open_File (buf, flags, 0);
			/* No GC has been taken place in Open_File() if it returns Null.
			 */
			if (!Nullp (port)) {
				Alloca_End;
				return port;
			}
		}
	}
	if (gotpath)
		Primitive_Error (x_("file ~s not found"), name);
	if (len + 1 > blen)
		Alloca (buf, CHAR*, len + 1);
	bcopy (fn, buf, len);
	buf[len] = '\0';
	port = Open_File (buf, flags, 1);
	Alloca_End;
	return port;
}

ELKObject P_Open_Input_File (ELKObject name) {
	return General_Open_File (name, P_INPUT, Null);
}

ELKObject P_Open_Output_File (ELKObject name) {
	return General_Open_File (name, 0, Null);
}

ELKObject P_Open_Input_Output_File (ELKObject name) {
	return General_Open_File (name, P_BIDIR, Null);
}

ELKObject General_Close_Port (ELKObject port) {
	register int flags, err = 0;
	FILE *f;

	Check_Type (port, T_Port);
	flags = PORT(port)->flags;
	if (!(flags & P_OPEN) || (flags & P_STRING))
		return Void;
	f = PORT(port)->file;
	if (f == stdin || f == stdout)
		return Void;
	if ((PORT(port)->closefun) (f) == EOF) {
		Saved_Errno = errno;   /* errno valid here? */
		err++;
	}
	PORT(port)->flags &= ~P_OPEN;
	Deregister_Object (port);
	if (err)
		Primitive_Error (x_("write error on ~s: ~E"), port);
	return Void;
}

ELKObject P_Close_Input_Port (ELKObject port) {
	return General_Close_Port (port);
}

ELKObject P_Close_Output_Port (ELKObject port){
	return General_Close_Port (port);
}

#define General_With(prim,curr,flags) ELKObject prim (ELKObject name, ELKObject thunk) {\
	ELKObject old, ret;\
	GC_Node2;\
\
	Check_Procedure (thunk);\
	old = curr;\
	GC_Link2 (thunk, old);\
	curr = General_Open_File (name, flags, Null);\
	ret = Funcall (thunk, Null, 0);\
	(void)General_Close_Port (curr);\
	GC_Unlink;\
	curr = old;\
	return ret;\
}

General_With (P_With_Input_From_File, Curr_Input_Port, P_INPUT)
General_With (P_With_Output_To_File, Curr_Output_Port, 0)

ELKObject General_Call_With (ELKObject name, int flags, ELKObject proc) {
	ELKObject port, ret;
	GC_Node2;

	Check_Procedure (proc);
	GC_Link2 (proc, port);
	port = General_Open_File (name, flags, Null);
	port = Cons (port, Null);
	ret = Funcall (proc, port, 0);
	(void)General_Close_Port (Car (port));
	GC_Unlink;
	return ret;
}

ELKObject P_Call_With_Input_File (ELKObject name, ELKObject proc) {
	return General_Call_With (name, P_INPUT, proc);
}

ELKObject P_Call_With_Output_File (ELKObject name, ELKObject proc) {
	return General_Call_With (name, 0, proc);
}

ELKObject P_Open_Input_String (ELKObject string) {
	Check_Type (string, T_String);
	return Make_Port (P_STRING|P_INPUT, (FILE *)0, string);
}

ELKObject P_Open_Output_String (void) {
	return Make_Port (P_STRING, (FILE *)0, Make_String ((CHAR *)0, 0));
}
