/* -*- tab-width: 4 -*- */

/* Output functions and primitives.
 */

#include "elkkernel.h"

#include <errno.h>
#include <ctype.h>
#ifdef STDC_HEADERS
#  include <stdarg.h>
#  define var_start(x, y) va_start(x, y)
#else
#  include <varargs.h>
#  define var_start(x, y) va_start(x)
#endif

#ifdef FLUSH_TIOCFLUSH
#  include <sys/ioctl.h>
#else
#  ifdef FLUSH_TCFLSH
#    include <termio.h>
#  endif
#endif

#ifdef ONUNIX
extern int errno;
#endif

int Saved_Errno;

static ELKObject V_Print_Depth, V_Print_Length;
void Pr_List (ELKObject port, ELKObject list, int raw, int depth, int length);
void Pr_Vector (ELKObject port, ELKObject vec, int raw, int depth, int length);
void Pr_Symbol (ELKObject port, ELKObject sym, int raw);
void Pr_String (ELKObject port, ELKObject str, int raw);
void Flush_Output (ELKObject port);
void DumpStdOut(CHAR *str, long len);
int Print_Length (void);
int Print_Depth (void);
void Print_Char (ELKObject port, int c);
void Print_String (ELKObject port, CHAR *buf, int len);
void Pr_Char (ELKObject port, int c);
void Print_Special (ELKObject port, int c);
ELKObject General_Print (int argc, ELKObject *argv, int raw);

void Init_Print (void) {
	Define_Variable (&V_Print_Depth, x_("print-depth"),
		Make_Integer (DEF_PRINT_DEPTH));
	Define_Variable (&V_Print_Length, x_("print-length"),
		Make_Integer (DEF_PRINT_LEN));
}

int Print_Length (void) {
	ELKObject pl;

	pl = Var_Get (V_Print_Length);
	return TYPE(pl) == T_Fixnum ? FIXNUM(pl) : DEF_PRINT_LEN;
}

int Print_Depth (void) {
	ELKObject pd;

	pd = Var_Get (V_Print_Depth);
	return TYPE(pd) == T_Fixnum ? FIXNUM(pd) : DEF_PRINT_DEPTH;
}

void Print_Char (ELKObject port, int c)
{
	CHAR buf[1];

	if (PORT(port)->flags & P_STRING)
	{
		buf[0] = c;
		Print_String (port, buf, 1);
	} else
	{
		FILE *io;

		io = PORT(port)->file;
		if (io == stdout || io == stderr)
		{
			CHAR cs[1];
			cs[0] = c;
			DumpStdOut(cs, 1);
		} else
		{
			if (putc (c, io) == EOF)
			{
				Saved_Errno = errno;   /* errno valid here? */
				Primitive_Error (x_("write error on ~s: ~E"), port);
			}
		}
	}
}

void Print_String (ELKObject port, register CHAR *buf, register int len) {
	register int n;
	register struct S_Port *p;
	ELKObject newo;
	GC_Node;

	p = PORT(port);
	n = STRING(p->name)->size - p->ptr;
	if (n < len) {
		GC_Link (port);
		n = len - n;
		if (n < STRING_GROW_SIZE)
			n = STRING_GROW_SIZE;
		newo = Make_String ((CHAR *)0, STRING(p->name)->size + n);
		p = PORT(port);
		GC_Unlink;
		bcopy (STRING(p->name)->data, STRING(newo)->data, p->ptr);
		p->name = newo;
	}
	bcopy (buf, STRING(p->name)->data + p->ptr, len);
	p->ptr += len;
}

/*VARARGS0*/
void Printf (ELKObject port, CHAR *fmt, ...)
{
	va_list args;
	CHAR buf[1024];

	var_start (args, fmt);
	if (PORT(port)->flags & P_STRING)
	{
		evsprintf (buf, fmt, args);
		Print_String (port, buf, estrlen (buf));
	} else
	{
		FILE *io;

		io = PORT(port)->file;
		if (io == stdout || io == stderr)
		{
			evsprintf (buf, fmt, args);
			DumpStdOut(buf, estrlen(buf));
		} else
		{
			evfprintf (io, fmt, args);
			if (ferror (io))
			{
				Saved_Errno = errno;   /* errno valid here? */
				Primitive_Error (x_("write error on ~s: ~E"), port);
			}
		}
	}
	va_end (args);
}

ELKObject General_Print (int argc, ELKObject *argv, int raw) {
	General_Print_Object (argv[0], argc == 2 ? argv[1] : Curr_Output_Port, raw);
	return Void;
}

ELKObject P_Write (int argc, ELKObject *argv) {
	return General_Print (argc, argv, 0);
}

ELKObject P_Display (int argc, ELKObject *argv) {
	return General_Print (argc, argv, 1);
}

ELKObject P_Write_Char (int argc, ELKObject *argv) {
	Check_Type (argv[0], T_Character);
	return General_Print (argc, argv, 1);
}

/*VARARGS1*/
ELKObject P_Newline (int argc, ELKObject *argv) {
	General_Print_Object (Newline, argc == 1 ? argv[0] : Curr_Output_Port, 1);
	return Void;
}

ELKObject P_Print (int argc, ELKObject *argv) {
	ELKObject port;
	GC_Node;

	port = argc == 2 ? argv[1] : Curr_Output_Port;
	GC_Link (port);
	General_Print_Object (argv[0], port, 0);
	Print_Char (port, '\n');
	Flush_Output (port);
	GC_Unlink;
	return Void;
}

ELKObject P_Clear_Output_Port (int argc, ELKObject *argv) {
	Discard_Output (argc == 1 ? argv[0] : Curr_Output_Port);
	return Void;
}

void Discard_Output (ELKObject port) {
	register FILE *f;

	Check_Output_Port (port);
	if (PORT(port)->flags & P_STRING)
		return;
	f = PORT(port)->file;
#ifdef FLUSH_FPURGE
	(void)fpurge (f);
#else
#ifdef FLUSH_BSD
	f->_cnt = 0;
	f->_ptr = f->_base;
#endif
#endif
#ifdef FLUSH_TIOCFLUSH
	(void)ioctl (fileno (f), TIOCFLUSH, (CHAR *)0);
#else
#ifdef FLUSH_TCFLSH
	(void)ioctl (fileno (f), TCFLSH, (CHAR *)1);
#endif
#endif
}

ELKObject P_Flush_Output_Port (int argc, ELKObject *argv) {
	Flush_Output (argc == 1 ? argv[0] : Curr_Output_Port);
	return Void;
}

void Flush_Output (ELKObject port)
{
	FILE *io;

	Check_Output_Port (port);
	if (PORT(port)->flags & P_STRING)
		return;
	io = PORT(port)->file;
	if (io == stderr || io == stdout) return;
	if (fflush (io) == EOF)
	{
		Saved_Errno = errno;   /* errno valid here? */
		Primitive_Error (x_("write error on ~s: ~E"), port);
	}
}

ELKObject P_Get_Output_String (ELKObject port) {
	register struct S_Port *p;
	ELKObject str;
	GC_Node;

	Check_Output_Port (port);
	GC_Link (port);
	str = Make_String ((CHAR *)0, PORT(port)->ptr);
	p = PORT(port);
	bcopy (STRING(p->name)->data, STRING(str)->data, p->ptr);
	p->ptr = 0;
	GC_Unlink;
	return str;
}

void Check_Output_Port (ELKObject port) {
	Check_Type (port, T_Port);
	if (!(PORT(port)->flags & P_OPEN))
		Primitive_Error (x_("port has been closed: ~s"), port);
	if (!IS_OUTPUT(port))
		Primitive_Error (x_("not an output port: ~s"), port);
}

void General_Print_Object (ELKObject x, ELKObject port, int raw) {
	Check_Output_Port (port);
	Print_Object (x, port, raw, Print_Depth (), Print_Length ());
}

void Print_Object (ELKObject x, ELKObject port, int raw, int depth, int length) {
	register int t;
	GC_Node2;

	GC_Link2 (port, x);
	t = TYPE(x);
	switch (t) {
	case T_Null:
		Printf (port, x_("()"));
		break;
	case T_Fixnum:
		Printf (port, x_("%d"), FIXNUM(x));
		break;
	case T_Bignum:
		Print_Bignum (port, x);
		break;
	case T_Flonum:
		Printf (port, x_("%s"), Flonum_To_String (x));
		break;
	case T_Boolean:
		Printf (port, x_("#%c"), FIXNUM(x) ? 't' : 'f');
		break;
	case T_Unbound:
		Printf (port, x_("#[unbound]"));
		break;
	case T_Special:
		Printf (port, x_("#[special]"));
		break;
	case T_Character: {
		int c = ELKCHAR(x);
		if (raw)
			Print_Char (port, c);
		else
			Pr_Char (port, c);
		break;
	}
	case T_Symbol:
		Pr_Symbol (port, x, raw);
		break;
	case T_Pair:
		Pr_List (port, x, raw, depth, length);
		break;
	case T_Environment:
		Printf (port, x_("#[environment %lu]"), POINTER(x));
		break;
	case T_String:
		Pr_String (port, x, raw);
		break;
	case T_Vector:
		Pr_Vector (port, x, raw, depth, length);
		break;
	case T_Primitive:
		Printf (port, x_("#[primitive %s]"), PRIM(x)->name);
		break;
	case T_Compound:
		if (Nullp (COMPOUND(x)->name)) {
			Printf (port, x_("#[compound %lu]"), POINTER(x));
		} else {
			Printf (port, x_("#[compound "));
			Print_Object (COMPOUND(x)->name, port, raw, depth, length);
			Print_Char (port, ']');
		}
		break;
	case T_Control_Point:
		Printf (port, x_("#[control-point %lu]"), POINTER(x));
		break;
	case T_Promise:
		Printf (port, x_("#[promise %lu]"), POINTER(x));
		break;
	case T_Port: {
		int str = PORT(x)->flags & P_STRING;
		CHAR *p;
		switch (PORT(x)->flags & (P_INPUT|P_BIDIR)) {
		case 0:	   p = x_("output");	   break;
		case P_INPUT: p = x_("input");		break;
		default:	  p = x_("input-output"); break;
		}
		Printf (port, x_("#[%s-%s-port "), str ? x_("string") : x_("file"), p);
		if (str)
			Printf (port, x_("%lu"), POINTER(x));
		else
			Pr_String (port, PORT(x)->name, 0);
		Print_Char (port, ']');
		break;
	}
	case T_End_Of_File:
		Printf (port, x_("#[end-of-file]"));
		break;
	case T_Autoload:
		Printf (port, x_("#[autoload "));
		Print_Object (AUTOLOAD(x)->files, port, raw, depth, length);
		Print_Char (port, ']');
		break;
	case T_Macro:
		if (Nullp (MACRO(x)->name)) {
			Printf (port, x_("#[macro %lu]"), POINTER(x));
		} else {
			Printf (port, x_("#[macro "));
			Print_Object (MACRO(x)->name, port, raw, depth, length);
			Print_Char (port, ']');
		}
		break;
	case T_Broken_Heart:
		Printf (port, x_("!!broken-heart!!"));
		break;
	default:
		if (t < 0 || t >= Num_Types)
			Panic (x_("bad type in print"));
		(Types[t].print)(x, port, raw, depth, length);
	}
	GC_Unlink;
}

void Pr_Char (ELKObject port, register int c) {
	register CHAR *p = 0;

	switch (c) {
	case ' ':
		p = x_("#\\space");
		break;
	case '\t':
		p = x_("#\\tab");
		break;
	case '\n':
		p = x_("#\\newline");
		break;
	case '\r':
		p = x_("#\\return");
		break;
	case '\f':
		p = x_("#\\formfeed");
		break;
	case '\b':
		p = x_("#\\backspace");
		break;
	default:
		if (c > ' ' && c < '\177')
			Printf (port, x_("#\\%c"), c);
		else
			Printf (port, x_("#\\%03o"), (UCHAR1)c);
	}
	if (p) Printf (port, p);
}

void Pr_List (ELKObject port, ELKObject list, int raw, int depth, int length) {
	ELKObject tail;
	register int len;
	register CHAR *s = 0;
	GC_Node2;

	if (depth == 0) {
		Printf (port, x_("&"));
		return;
	}
	GC_Link2 (port, list);
	if (!Nullp (list) && ((tail = Cdr (list)), TYPE(tail) == T_Pair)
					  && ((tail = Cdr (tail)), Nullp (tail))) {
		tail = Car (list);
		if (EQ(tail, Sym_Quote))
			s = x_("'");
		else if (EQ(tail, Sym_Quasiquote))
			s = x_("`");
		else if (EQ(tail, Sym_Unquote))
			s = x_(",");
		else if (EQ(tail, Sym_Unquote_Splicing))
			s = x_(",@");
		if (s) {
			Printf (port, s);
			Print_Object (Car (Cdr (list)), port, raw,
				depth < 0 ? depth : depth-1, length);
			GC_Unlink;
			return;
		}
	}
	Print_Char (port, '(');
	for (len = 0; !Nullp (list); len++, list = tail) {
		if (length >= 0 && len >= length) {
			Printf (port, x_("..."));
			break;
		}
		Print_Object (Car (list), port, raw, depth < 0 ? depth : depth-1,
			 length);
		tail = Cdr (list);
		if (!Nullp (tail)) {
			if (TYPE(tail) == T_Pair)
				Print_Char (port, ' ');
			else {
				Printf (port, x_(" . "));
				Print_Object (tail, port, raw, depth < 0 ? depth : depth-1,
					 length);
				break;
			}
		}
	}
	Print_Char (port, ')');
	GC_Unlink;
}

void Pr_Vector (ELKObject port, ELKObject vec, int raw, int depth, int length) {
	register int i, j;
	GC_Node2;

	if (depth == 0) {
		Printf (port, x_("&"));
		return;
	}
	GC_Link2 (port, vec);
	Printf (port, x_("#("));
	for (i = 0, j = VECTOR(vec)->size; i < j; i++) {
		if (i) Print_Char (port, ' ');
		if (length >= 0 && i >= length) {
			Printf (port, x_("..."));
			break;
		}
		Print_Object (VECTOR(vec)->data[i], port, raw,
			 depth < 0 ? depth : depth-1, length);
	}
	Print_Char (port, ')');
	GC_Unlink;
}

void Pr_Symbol (ELKObject port, ELKObject sym, int raw) {
	ELKObject str;
	register int c, i;
	GC_Node2;

	str = SYMBOL(sym)->name;
	if (raw) {
		Pr_String (port, str, raw);
		return;
	}
	GC_Link2 (port, str);
	for (i = 0; i < STRING(str)->size; i++) {
		c = STRING(str)->data[i];
		switch (c) {
		case '\\': case ';': case '#': case '(': case ')':
		case '\'': case '`': case ',': case '"': case '.':
		case '\t': case '\n': case ' ':
			Print_Char (port, '\\');
			Print_Char (port, c);
			break;
		default:
			if (c < ' ' || c >= '\177')
				Print_Special (port, c);
			else
				Print_Char (port, c);
		}
	}
	GC_Unlink;
}

void Pr_String (ELKObject port, ELKObject str, int raw)
{
	register CHAR *p = STRING(str)->data;
	register int c, i, len = STRING(str)->size;
	GC_Node2;

	if (raw)
	{
		if (PORT(port)->flags & P_STRING)
		{
			Print_String (port, p, len);
		} else
		{
			FILE *io;
			io = PORT(port)->file;
			if (io == stdout || io == stderr)
			{
				DumpStdOut(p, len);
			} else
			{
				if ((int)fwrite (p, 1, len, io) != len)
				{
					Saved_Errno = errno;   /* errno valid here? */
					Primitive_Error (x_("write error on ~s: ~E"), port);
				}
			}
		}
		return;
	}
	GC_Link2 (port, str);
	Print_Char (port, '"');
	for (i = 0; i < STRING(str)->size; i++) {
		c = STRING(str)->data[i];
		if (c == '\\' || c == '"')
			Print_Char (port, '\\');
		if (c < ' ' || c >= '\177')
			Print_Special (port, c);
		else
			Print_Char (port, c);
	}
	Print_Char (port, '"');
	GC_Unlink;
}

void Print_Special (ELKObject port, register int c) {
	register CHAR *fmt = x_("\\%c");

	switch (c) {
	case '\b': c = 'b'; break;
	case '\t': c = 't'; break;
	case '\r': c = 'r'; break;
	case '\n': c = 'n'; break;
	default:
		fmt = x_("\\%03o");
	}
	Printf (port, fmt, (UCHAR1)c);
}

ELKObject P_Format (int argc, ELKObject *argv) {
	ELKObject port, str;
	register int stringret = 0;
	GC_Node;

	port = argv[0];
	if (TYPE(port) == T_Boolean) {
		if (Truep (port)) {
			port = Curr_Output_Port;
		} else {
			stringret++;
			port = P_Open_Output_String ();
		}
	} else if (TYPE(port) == T_Port) {
		Check_Output_Port (port);
	} else Wrong_Type_Combination (port, x_("port or #t or #f"));
	str = argv[1];
	Check_Type (str, T_String);
	GC_Link (port);
	Format (port, STRING(str)->data, STRING(str)->size, argc-2, argv+2);
	GC_Unlink;
	return stringret ? P_Get_Output_String (port) : Void;
}

void Format (ELKObject port, const CHAR *fmt, int len, int argc, ELKObject *argv) {
	register const CHAR *ep;
	CHAR *p;
	register int c;
	CHAR buf[256];
	GC_Node;
	Alloca_Begin;

	GC_Link (port);
	Alloca (p, CHAR*, len);
	bcopy (fmt, p, len);
	for (ep = p + len; p < ep; p++) {
		if (*p == '~') {
			if (++p == ep) break;
			if ((c = *p) == '~') {
				Print_Char (port, c);
			} else if (c == '%') {
				Print_Char (port, '\n');
			} else if (c == 'e' || c == 'E') {
				esnprintf (buf, 256, x_("error %d"), Saved_Errno);
				Print_Object (Make_String (buf, estrlen (buf)), port,
					c == 'E', 0, 0);
			} else {
				if (--argc < 0)
					Primitive_Error (x_("too few arguments"));
				if (c == 's' || c == 'a') {
					Print_Object (*argv, port, c == 'a', Print_Depth (),
						Print_Length ());
					argv++;
				} else if (c == 'c') {
					Check_Type (*argv, T_Character);
					Print_Char (port, ELKCHAR(*argv));
					argv++;
				} else Print_Char (port, c);
			}
		} else {
			Print_Char (port, *p);
		}
	}
	Alloca_End;
	GC_Unlink;
}
