/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

/*
 * The routines in this file handle binary network files input/output.
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "irsimnet_macros.h"


static	FILE	*fnet;			/* file for reading/writing network */
static	tptr	rd_tlist;		/* list of transistor just read */
static	double	inet_CTGA;		/* transistor capacitances */
static	double	inet_CTDE;
static	double	inet_CTDW;
static	long	inet_lambda;
static	void	adjust_transistors(tptr tlist);

static	CHAR	netHeader[] = {x_("<<inet>>")};
static	CHAR	endAscii[] = {x_("<end>\n")};

#define	CM_M	100.0			/* centimicrons per micron */
#define	CM_M2	(CM_M * CM_M)		/* centimicrons per micron ^ 2 */


#define	HEADER_LEN		(sizeof(netHeader) - 1)

	/* define number of bytes used to write/read network parameter */
#define	NB_NDINDEX	4
#define	NB_NCAP		4
#define	NB_TSIZE	4
#define	NB_COORD	4
#define	NB_VTHRESH	2
#define	NB_TP		2
#define	NB_STRLEN	2
#define	NB_FLAGS	1
#define	NB_TTYPE	1

	/* node flags we are interested in preserving */
#define	SAVE_FLAGS	(POWER_RAIL | ALIAS | USERDELAY)

	/* scale factors for floating-point numbers */
#define	VTH_SCALE		1000.0
#define	CAP_SCALE		16384.0    /* power of 2 to minimize error */


typedef struct		/* format of node records in net file */
{
	union
	{
		CHAR  cap[NB_NCAP];
		CHAR  alias[NB_NDINDEX];
	} n;
	CHAR  flags[NB_FLAGS];
	CHAR  slen[NB_STRLEN];
} File_Node;

#define	Size_File_Node		(NB_NDINDEX + NB_FLAGS + NB_STRLEN)


typedef struct		/* format of node user-delays in net file */
{
	CHAR  tphl[NB_TP];
	CHAR  tplh[NB_TP];
} File_Delay;

#define	Size_File_Delay		(NB_TP + NB_TP)


typedef struct		/* format of node thresholds in net file */
{
	CHAR  vhigh[NB_VTHRESH];
	CHAR  vlow[NB_VTHRESH];
} File_Thresh;

#define	Size_File_Thresh 	(NB_VTHRESH + NB_VTHRESH)


typedef struct		/* format of transistor records in net file */
{
	CHAR  width[NB_TSIZE];
	CHAR  length[NB_TSIZE];
	CHAR  ttype[NB_TTYPE];
	CHAR  gate[NB_NDINDEX];
	CHAR  src[NB_NDINDEX];
	CHAR  drn[NB_NDINDEX];
	CHAR  x[NB_COORD];
	CHAR  y[NB_COORD];
} File_Trans;

#define	Size_File_Trans	(NB_TSIZE + NB_TSIZE + NB_TTYPE + \
		NB_NDINDEX + NB_NDINDEX + NB_NDINDEX + NB_COORD + NB_COORD)


static tptr wr_nodes(void)
{
	register nptr  n;
	register tptr  txlist;
	register lptr  l;
	File_Node      node;
	File_Delay     delays;
	File_Thresh    thresh;
	int            slen;
	double         cap;
	float          vhigh = (float)irsim_HIGHTHRESH, vlow = (float)irsim_LOWTHRESH;

	txlist = NULL;

	for(n = irsim_GetNodeList(); n != NULL; n = n->n.next)
	{
		cap = n->ncap;
		for(l = n->nterm; l != NULL; l = l->next)
		{
			register tptr  t = l->xtor;

			if (! (t->tflags & ACTIVE_T))
			{
				t->scache.t = txlist;
				txlist = t;
				t->tflags |= ACTIVE_T;
			}
			if (t->ttype & GATELIST)
				cap += (irsim_StackCap(t) / 2.0);
		}

		slen = estrlen(pnode(n)) + 1;
		irsim_PackBytes(node.slen, slen, NB_STRLEN);
		node.flags[0] = (CHAR)(n->nflags & SAVE_FLAGS);
		if (n->vhigh != vhigh || n->vlow != vlow)
			node.flags[0] |= WATCHED;
		if (n->nflags & ALIAS)
			irsim_PackBytes(node.n.alias, irsim_Node2index(n->nlink), NB_NDINDEX);
		else
			irsim_PackBytes(node.n.cap, castint((float)(cap * CAP_SCALE)), NB_NCAP);
		(void)xfwrite((UCHAR1 *)&node, 1, Size_File_Node, fnet);
		(void)xfwrite((UCHAR1 *)pnode(n), 1, slen, fnet);

		if (n->nflags & USERDELAY)
		{
			irsim_PackBytes(delays.tphl, n->tphl, NB_TP);
			irsim_PackBytes(delays.tplh, n->tplh, NB_TP);
			(void)xfwrite((UCHAR1 *)&delays, 1, Size_File_Delay, fnet);
		}
		if (node.flags[0] & WATCHED)
		{
			irsim_PackBytes(thresh.vhigh, castint((float)(n->vhigh * VTH_SCALE)), NB_VTHRESH);
			irsim_PackBytes(thresh.vlow, castint((float)(n->vlow * VTH_SCALE)), NB_VTHRESH);
			(void)xfwrite((UCHAR1 *)&thresh, 1, Size_File_Thresh, fnet);
		}
	}

	return(txlist);
}


static void wr_trans(register tptr t)
{
	File_Trans  trans;
	Ulong       ndindx;

	if (t->ttype & ORED)
	{
		for(t = t->tlink; t != NULL; t = t->scache.t) wr_trans(t);
		return;
	}

	ndindx = irsim_Node2index(t->gate);
	irsim_PackBytes(trans.gate, ndindx, NB_NDINDEX);

	ndindx = irsim_Node2index(t->source);
	irsim_PackBytes(trans.src, ndindx, NB_NDINDEX);

	ndindx = irsim_Node2index(t->drain);
	irsim_PackBytes(trans.drn, ndindx, NB_NDINDEX);

	irsim_PackBytes(trans.length, t->r->length, NB_TSIZE);
	irsim_PackBytes(trans.width, t->r->width, NB_TSIZE);

	trans.ttype[0] = t->ttype & ~(ORED | ORLIST | STACKED | GATELIST);

	if (t->tlink != t)
	{
		trans.ttype[0] |= GATELIST;
		irsim_PackBytes(trans.x, t->x.pos, NB_COORD);
		irsim_PackBytes(trans.y, t->y.pos, NB_COORD);
	}

	(void)xfwrite((UCHAR1 *)&trans, 1, Size_File_Trans, fnet);
}


static void wr_txtors(tptr tlist)
{
	register tptr  t, next;

	for(t = tlist; t != NULL; t = next)
	{
		next = t->scache.t;

		t->tflags &= ~ACTIVE_T;
		t->scache.t = NULL;

		if (t->ttype & GATELIST)
		{
			for(t = (tptr) t->gate; t != NULL; t = t->scache.t)
				wr_trans(t);
		}
		else
			wr_trans(t);
	}

	for(t = irsim_tcap->dcache.t; t != irsim_tcap; t = t->dcache.t)
		wr_trans(t);
}


static void WriteAscii(FILE  *f)
{
	(void)efprintf(f, x_("lambda %ld\n"), irsim_LAMBDACM);
	(void)efprintf(f, x_("CTGA %.6f\n"), irsim_CTGA * CM_M2);
	(void)efprintf(f, x_("CTDW %.6f\n"), irsim_CTDW * CM_M);
	(void)efprintf(f, x_("CTDE %.6f\n"), irsim_CTDE);

	/* add extensions here */
	(void)efprintf(f, endAscii);
}


static int ReadAscii(FILE  *f, CHAR  *line)
{
	int  nl = 0;

	inet_CTGA = irsim_CTGA;
	inet_CTDE = irsim_CTDE;
	inet_CTDW = irsim_CTDW;
	inet_lambda = irsim_LAMBDACM;

	do
	{
		if (xfgets(line, 200, f))
			return(-1);
		nl++;
		if (estrcmp(line, endAscii) == 0)
			return(0);
		if (estrncmp(line, x_("lambda "), 7) == 0)
			inet_lambda = eatoi(line + 7);
		if (estrncmp(line, x_("CTGA "), 5) == 0)
			inet_CTGA = eatof(line + 5) / CM_M2;
		else if (estrncmp(line, x_("CTDW "), 5) == 0)
			inet_CTDW = eatof(line + 5) / CM_M;
		else if (estrncmp(line, x_("CTDE "), 5) == 0)
			inet_CTDE = eatof(line + 5);

		/* add extensions here */
	}
	while(nl < 30);

	ttyputerr(x_("inet file seems bad"));
	return(-1);
}


void irsim_wr_netfile(CHAR *fname)
{
	tptr  txlist;

	fnet = xcreate(fname, el_filetypetext, 0, 0);
	if (fnet == NULL)
	{
		ttyputerr(x_("can't open file '%s'"), fname);
		return;
	}

	(void)efprintf(fnet, x_("%s\n"), netHeader);
	(void)efprintf(fnet, x_("%d %d\n"), irsim_GetHashSize(), irsim_nnodes + irsim_naliases);

	WriteAscii(fnet);
	txlist = wr_nodes();
	wr_txtors(txlist);
	(void)fclose(fnet);
}


static void PrematureEof(void)
{
	ttyputerr(x_("premature eof in inet file"));
}


static int rd_nodes(CHAR  *nname, int n_nodes)
{
	File_Node    node;
	File_Delay   delays;
	File_Thresh  thresh;
	Ulong        tmp;
	int          slen;
	nptr         n;
	nptr         aliases;

	aliases = NULL;
	while(n_nodes-- != 0)
	{
		if (xfread((UCHAR1 *)&node, 1, Size_File_Node, fnet) != Size_File_Node)
		{
			PrematureEof();
			return(FALSE);
		}

		slen = (int)irsim_UnpackBytes(node.slen, NB_STRLEN);
		if (xfread((UCHAR1 *)nname, 1, slen, fnet) != slen)
		{
			PrematureEof();
			return(FALSE);
		}

		n = irsim_GetNode(nname);

		n->nflags = node.flags[0];
		if (n->nflags & ALIAS)
		{
			tmp = (Ulong)irsim_UnpackBytes(node.n.alias, NB_NDINDEX);
			n->c.nindex = tmp;

			n->n.next = aliases;
			aliases = n;
		} else
		{
			tmp = (Ulong)irsim_UnpackBytes(node.n.cap, NB_NCAP);
			n->ncap = (float)(tmp * (1/CAP_SCALE));
			if (n->ncap < MIN_CAP)
				n->ncap = (float)MIN_CAP;
		}

		if (n->nflags & USERDELAY)
		{
			if (xfread((UCHAR1 *)&delays, 1, Size_File_Delay, fnet) != Size_File_Delay)
			{
				PrematureEof();
				return(FALSE);
			}
			n->tphl = (short)irsim_UnpackBytes(delays.tphl, NB_TP);
			n->tplh = (short)irsim_UnpackBytes(delays.tplh, NB_TP);
		}
		if (n->nflags & WATCHED)
		{
			n->nflags &= ~WATCHED;
			if (xfread((UCHAR1 *)&thresh, 1, Size_File_Thresh, fnet) != Size_File_Thresh)
			{
				PrematureEof();
				return(FALSE);
			}
			tmp = (Ulong)irsim_UnpackBytes(thresh.vhigh, NB_VTHRESH);
			n->vhigh = (float)(tmp * (1/VTH_SCALE));
			tmp = (Ulong)irsim_UnpackBytes(thresh.vlow, NB_VTHRESH);
			n->vlow = (float)(tmp * (1/VTH_SCALE));
		}
	}

	irsim_VDD_node->nflags |= INPUT;		/* these bits get thrashed above */
	irsim_GND_node->nflags |= INPUT;

	for(n = aliases; n != NULL; n = n->n.next)
	{
		n->nlink = irsim_Index2node(n->c.nindex);
		n->c.nindex = 0;
		irsim_nnodes--;
		irsim_naliases++;
	}
	return(TRUE);
}


static void rd_txtors(void)
{
	File_Trans  trans;
	Ulong       ndindx;
	long        width, length;
	tptr        t, *last;

	rd_tlist = NULL;
	last = &rd_tlist;
	while(xfread((UCHAR1 *)&trans, 1, Size_File_Trans, fnet) == Size_File_Trans)
	{
		NEW_TRANS(t);

		ndindx = (Ulong)irsim_UnpackBytes(trans.gate, NB_NDINDEX);
		t->gate = irsim_Index2node(ndindx);

		ndindx = (Ulong)irsim_UnpackBytes(trans.src, NB_NDINDEX);
		t->source = irsim_Index2node(ndindx);

		ndindx = (Ulong)irsim_UnpackBytes(trans.drn, NB_NDINDEX);
		t->drain = irsim_Index2node(ndindx);

		width = (long)irsim_UnpackBytes(trans.width, NB_TSIZE);
		length = (long)irsim_UnpackBytes(trans.length, NB_TSIZE);

		t->ttype = trans.ttype[0];

		if (t->ttype & GATELIST)
		{
			t->ttype &= ~GATELIST;
			t->x.pos = (long)irsim_UnpackBytes(trans.x, NB_COORD);
			t->y.pos = (long)irsim_UnpackBytes(trans.y, NB_COORD);
			irsim_EnterPos(t, TRUE);
		} else
			irsim_EnterPos(t, FALSE);

		t->r = irsim_requiv((int) t->ttype, width, length);

		irsim_ntrans[BASETYPE(t->ttype)] += 1;

		*last = t;
		last = &(t->scache.t);
	}
	*last = NULL;
}


int irsim_rd_netfile(FILE *f, CHAR *line)
{
	int   hash_size, n_nodes;

	if (estrncmp(line, netHeader, HEADER_LEN) != 0)
		return(FALSE);

	if (xfgets(line, 200, f))
	{
		PrematureEof();
		return(FALSE);
	}

	if (esscanf(line, x_("%d %d"), &hash_size, &n_nodes) != 2)
	{
		ttyputerr(x_("bad format for net file"));
		return(FALSE);
	}

	if (hash_size != irsim_GetHashSize())
	{
		ttyputerr(x_("Incompatible net file version"));
		return(FALSE);
	}

	if (ReadAscii(f, line))
	{
		PrematureEof();
		return(FALSE);
	}

	fnet = f;
	if (!rd_nodes(line, n_nodes)) return(FALSE);
	rd_txtors();

	adjust_transistors(rd_tlist);

	return(TRUE);
}


static void adjust_transistors(tptr tlist)
{
	double         adj_g, adj_w, adj_e;
	register tptr  t;

#    define	ABS(X)	(((X) < 0) ? -(X) : (X))
#    ifndef CAP_ADJUST_LIMIT
#	define	CAP_ADJUST_LIMIT	0.0000005
#    endif
#    define	G_CAP_ADJUST_LIMIT	(CAP_ADJUST_LIMIT / CM_M2)
#    define	W_CAP_ADJUST_LIMIT	(CAP_ADJUST_LIMIT / CM_M)

	adj_g = irsim_CTGA - inet_CTGA;
	adj_w = irsim_CTDW - inet_CTDW;
	adj_e = irsim_CTDE - inet_CTDE;

	if (irsim_LAMBDACM != inet_lambda && irsim_LAMBDACM > 0 && inet_lambda > 0)
	{
		double  adj_size = (double) irsim_LAMBDACM / (double) inet_lambda;

		for(t = tlist; t != NULL; t = t->scache.t)
		{
			t->r->length = (Ulong)(t->r->length * adj_size);
			t->r->width = (Ulong)(t->r->width * adj_size);
		}
	}

	if (ABS(adj_g) > G_CAP_ADJUST_LIMIT)
	{
		for(t = tlist; t != NULL; t = t->scache.t)
			t->gate->ncap += (float)(adj_g * t->r->length * t->r->width);
	}

	if (ABS(adj_w) > W_CAP_ADJUST_LIMIT || ABS(adj_e) > CAP_ADJUST_LIMIT)
	{
		for(t = tlist; t != NULL; t = t->scache.t)
		{
			t->source->ncap += (float)(adj_w * t->r->width + adj_e);
			t->drain->ncap += (float)(adj_w * t->r->width + adj_e);
		}
	}
}


/*
 * Traverse the transistor list and add the node connection-list.  We have
 * to be careful with ALIASed nodes.  Note that transistors with source/drain
 * connected to VDD and GND nodes are not linked.
 */
nptr irsim_bin_connect_txtors(void)
{
	register tptr  t, tnext;
	register nptr  nd_list = NULL;

	for(t = rd_tlist; t != NULL; t = tnext)
	{
		tnext = t->scache.t;
		t->state = (t->ttype & ALWAYSON) ? WEAK : UNKNOWN;
		t->tflags = 0;

		if (t->ttype & TCAP)
		{
			LINK_TCAP(t);
		} else
		{
			if (t->ttype & ALWAYSON)
			{
				CONNECT(irsim_on_trans, t);
			} else
			{
				CONNECT(t->gate->ngate, t);
			}

			if (! (t->source->nflags & POWER_RAIL))
			{
				CONNECT(t->source->nterm, t);
				LINK_TO_LIST(t->source, nd_list, VISITED);
			}
			if (! (t->drain->nflags & POWER_RAIL))
			{
				CONNECT(t->drain->nterm, t);
				LINK_TO_LIST(t->drain, nd_list, VISITED);
			}
		}
	}
	return(nd_list);
}
