/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

/*
 * Incremental simulation
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"


/*
 * Set H to the next "effective" change in the history following Hp
 */
#define	NEXTH(H, Hp)					\
    for((H) = (Hp)->next; (H)->punt; (H) = (H)->next)


/*
 * time at which history entry was enqueued
 */
#define	QTIME(H)		((TimeType)(H)->htime - (TimeType)(H)->t.r.delay)
#define	QPTIME(H)		((TimeType)(H)->htime - (TimeType)(H)->t.p.delay)
#define	PuntTime(H)		((TimeType)(H)->htime - (TimeType)(H)->t.p.ptime)


#define	IsDeviated(ND)	\
    ((ND)->nflags & (DEVIATED | CHANGED) || (ND)->events != NULL)

#define	NeedUpdate(ND)	\
    (((ND)->nflags & (ACTIVE_CL | STIM | POWER_RAIL)) == 0)


long	irsim_INC_RES = 0;		/* time resolution in ticks (.5 ns) */
nptr	irsim_inc_cause = (nptr) 1;	/* fake cause ptr for path command */

long    irsim_nevals = 0;		/* # of evaluations */
long    irsim_i_nevals = 0;		/* # of incremental evaluations */

long	irsim_nreval_ev = 0;		/* number of various events */
long	irsim_npunted_ev = 0;
long	irsim_nstimuli_ev = 0;
long	irsim_ncheckpt_ev = 0;
long	irsim_ndelaychk_ev = 0;
long	irsim_ndelay_ev = 0;
int     irsim_fault_mode = FALSE;	/* TRUE when doing faultsim */

static	long    o_nevals, o_nevent;
static	void	(*curr_model)(nptr);	/* the current model used to eval */
static	hptr	modelp;			/* model history */

#ifdef FAULT_SIM
#  define TRIGGER_EV	0x81

  static int stop_early;		/* TRUE when fault detected */
  static TimeType old_cur_delta;	/* cur_delta before faultsim */
  static evptr pending_evs;		/* pending events before faultsim */

#  define DoingFault	(irsim_fault_mode != FALSE)
#  define NotDoingFault	(irsim_fault_mode == FALSE)
#else
#  define DoingFault	(0)
#  define NotDoingFault	(1)
#  define hchange		head
#endif


typedef struct			/* transistor stage built by GetConnList */
{
	int   flags;			/* see below */
	nptr  nd_list;			/* the connection list */
	nptr  inp_list;			/* input nodes connected to list */
} Stage, *pstg;

		/* stage flags */
#define	ALL_MERGED	0x1		/* all nodes in stage are converged */
#define	ONLY_INPUT	0x2		/* list consists of input node only */
#define	INP_TRANS	0x4		/* there are switching inputs */

void irsim_init_incsim(void)
{
	irsim_INC_RES = 0;
	irsim_inc_cause = (nptr) 1;
	irsim_nevals = 0;
	irsim_i_nevals = 0;
	irsim_nreval_ev = 0;
	irsim_npunted_ev = 0;
	irsim_nstimuli_ev = 0;
	irsim_ncheckpt_ev = 0;
	irsim_ndelaychk_ev = 0;
	irsim_ndelay_ev = 0;
	irsim_fault_mode = FALSE;
}

/*
 * Update the state of a node from the history.  Set the INPUT flag as well.
 * Return a pointer to the next change in the history.
 */
static	hptr UpdateNode(register nptr nd)
{
	register hptr cur, nxt;

	cur = nd->curr;
	if ((TimeType)cur->htime > irsim_cur_delta)
		cur = (hptr) &(nd->head);

	NEXTH(nxt, cur);
	while((TimeType)nxt->htime <= irsim_cur_delta)
	{
		cur = nxt;
		NEXTH(nxt, nxt);
	}

	nd->curr = cur;
	nd->npot = cur->val;
	if (cur->inp)
		nd->nflags |= INPUT;
	else
		nd->nflags &= ~INPUT;

	return(nxt);
}


/*
 * Activate a node: Set the ACTIVE_CL flag, schedule all pending transitions
 * as of irsim_cur_delta.  Punted events that were already punted (as of cur_delta)
 * remain in the history.  Pending punted events are rescheduled.  Punted
 * events between nd->curr and the next transition that were queued after
 * irsim_cur_delta are removed from the history and put in nd->t.punts.
 */
static void ActivateNode(nptr nd)
{
	register hptr  h, p;
	hptr           *lastp;
	int            first;

#ifdef INCDEBUG
	if (nd->nflags & WATCHED)
	ttyputmsg(x_("%.2f: Activate %s"), d2ns(irsim_cur_delta), pnode(nd));
#endif
	if (nd->nflags & STIM)
	{					/* don't dequeue if ntime == now */
		if (nd->c.event->ntime != irsim_cur_delta)
			irsim_DequeueEvent(nd);
		nd->nflags &= ~STIM;
		NEXTH(p, nd->curr);
	}
	else
		p = UpdateNode(nd);

	(void)irsim_EnqueueHist(nd, p, CHECK_PNT);
	nd->nflags |= (ACTIVE_CL | WAS_ACTIVE);

	lastp = &(nd->t.punts);
	*lastp = NULL;

	/* look for any pending events as of irsim_cur_delta */
	first = TRUE;
	for(p = nd->curr, h = p->next; ; p = h, h = h->next)
	{
		if (h->punt)
		{
			if (PuntTime(h) < irsim_cur_delta)	/* already punted, skip it */
				continue;

			if (QPTIME(h) <= irsim_cur_delta)	/* pending: enqueue it */
			{
				if (QPTIME(h) != irsim_cur_delta)	/* cur_delta: do not queue */
					(void)irsim_EnqueueHist(nd, h, PUNTED);
				if (DoingFault)
					continue;
				p->next = h->next;		/* and remove it from hist */
				h->next = irsim_freeHist;
				irsim_freeHist = h;
				h = p;
			} else if (first && NotDoingFault)
			{			/* move if punted before next transition */
				p->next = h->next;
				h->next = NULL;
				*lastp = h;
				lastp = &(h->next);
				h = p;
			}
		} else
		{
			first = FALSE;
			if (QTIME(h) < irsim_cur_delta)	/* pending: enqueue it */
				(void)irsim_EnqueueHist(nd, h, REVAL);
			else
				break;
		}
	}
}


/*
 * Define all possible cases due to an event on an active node, as follows:
 *	000 0: no edge, node remains converged (not deviated)
 *	001 1: no edge, node just converged
 *	010 2: no edge, node just deviated
 *	011 3: no edge, node remains deviated
 *	100 4: edge, node remains converged
 *	101 5: edge, node just converged
 *	110 6: edge, node just deviated
 *	111 7: edge, node remains deviated
 */

#define	O_DEV		0x1		/* node was previously deviated */
#define	N_DEV		0x2		/* node is now deviated */
#define	EDGE		0x4		/* a transistion ocurred */
#define	CHK		0x8		/* just dequeued a check_pnt event */
#define	SAME_T		0x10		/* event merges at the same time */
#define	N_ACTV		0x20		/* merge events: still active xtors */

#define	GetOdev(nd)		((nd)->nflags & DEVIATED)

static	evptr    dev_evs;		/* deviate events */
static	evptr    mrg_evs;		/* merge events w/transition */
static	evptr    *last_mrg;		/* last merge event w/transition */
static	evptr    mrg0_evs;		/* last merge event */
static	evptr    chk_evs;		/* events with no transition */
static	evptr    inp_evs;		/* input events */
static	evptr    xinp_evs;		/* stop being an input events */
static	evptr    stim_evs;		/* stimulus events */
static	evptr    pend_evs;		/* pending events */


/*
 * Change the event type to reflect new state, add event to its corresponding
 * event list, set the oldpot field and the DEVIATED bit for the node.
 */
#define	Deviate_Edge(ND, EV, Oval, Chk)		\
{							\
	(EV)->type = (UCHAR1)(GetOdev(ND) | N_DEV | Chk | EDGE);	\
	(EV)->nlink = dev_evs;				\
	dev_evs = (EV);					\
	(ND)->oldpot = (short)Oval;				\
	(ND)->nflags |= DEVIATED;				\
}							\


#define	Deviate(ND, EV, Oval, Chk)			\
{							\
	(EV)->type = (UCHAR1)(GetOdev(ND) | N_DEV | Chk);		\
	(EV)->nlink = chk_evs;				\
	chk_evs = (EV);					\
	(ND)->oldpot = Oval;				\
	(ND)->nflags |= DEVIATED;				\
}							\


#define	Deviate_SameVal(EV)				\
{							\
	(EV)->type = N_DEV | O_DEV | EDGE;			\
	(EV)->nlink = dev_evs;				\
	dev_evs = (EV);					\
}							\


/*
 * Change the event type to reflect new state, add event to its corresponding
 * event list and reset the DEVIATED bit.
 */
#define Merge_Edge(ND, EV, Chk)			\
{							\
	if ((ND)->nflags & DEVIATED)			\
	{							\
		(EV)->type = (O_DEV | EDGE | Chk);		\
		*last_mrg = (EV);				\
		last_mrg = & (EV)->nlink;			\
	} else						\
	{							\
		(EV)->type = (Chk | EDGE);			\
		(EV)->nlink = mrg0_evs;				\
		mrg0_evs = (EV);				\
	}							\
	(ND)->nflags &= ~DEVIATED;				\
}							\


#define Merge_SameEdge(ND, EV, Chk)			\
{							\
	(EV)->type = (UCHAR1)(GetOdev(ND) | SAME_T | Chk | EDGE);	\
	(EV)->nlink = mrg0_evs;				\
	mrg0_evs = (EV);					\
	(ND)->nflags &= ~DEVIATED;				\
}							\


#define Merge(ND, EV, Chk)				\
{							\
	(EV)->type = (UCHAR1)(GetOdev(ND) | Chk);			\
	(EV)->nlink = chk_evs;				\
	chk_evs = (EV);					\
	(ND)->nflags &= ~DEVIATED;				\
}							\


/*
 * Free all punted events currently in nd->t.punts and move all punted events
 * after nd->curr to nd->t.punts. 
 */
static void ReplacePunts(nptr nd)
{
	register hptr  h, p;

	if (DoingFault)
		return;
	h = nd->t.punts;
	if (h != NULL)
	{
		for(p = h; p->next != NULL; p = p->next);
			p->next = irsim_freeHist;
		irsim_freeHist = h;
	}

	h = nd->curr;
	for(p = h; p->next->punt; p = p->next);
	if (p->punt)
	{
		nd->t.punts = h->next;
		h->next = p->next;
		p->next = NULL;
	}
	else
		nd->t.punts = NULL;
}


/*
 * Update the state of a node due to a re-evaluation event.
 */
static void UpdateReval(register evptr e)
{
	register nptr   n;			/* node in question */
	register evptr  nxte;		/* next chk_point event (if any) */
	TimeType   d_t;		/* delta time between events */

	n = e->enode;
	nxte = n->c.event;

	/* no chk_point event or chk_point is too far in the future */
	if (nxte == NULL || (d_t = nxte->ntime - e->ntime) > irsim_INC_RES)
	{
		free_from_node(e, n);
		irsim_NewEdge(n, e);
		if (n->nflags & DEVIATED)
		{
			if (n->oldpot == e->eval)
				Merge_Edge(n, e, 0)
			else
				Deviate_SameVal(e)
		} else
			Deviate_Edge(n, e, n->npot, 0);
		n->npot = e->eval;
	} else if (d_t != 0)
	{
		free_from_node(e, n);

		switch(nxte->type)
		{
			case DELAY_CHK:
				irsim_ndelaychk_ev++;
				if (nxte->eval == e->eval)
				{
					n->curr = nxte->p.hist;	/* update curr pointer */
					Merge_Edge(n, e, CHK);
					ReplacePunts(n);
				} else
				{
					irsim_DeleteNextEdge(n);
					irsim_NewEdge(n, e);
					Deviate_Edge(n, e, nxte->eval, CHK);
				}
				irsim_DequeueEvent(n);
				n->npot = e->eval;
				break;

			case DELAY_EV:
				if (nxte->eval == e->eval)
					irsim_DelayEvent(e, d_t);		/* delay again */
				else
				{
					ttyputmsg(x_("Missed Glitch: %s => (%.2f %c) (%.2f %c)"),
						pnode(n), d2ns(nxte->p.oldt), x_("0XX1")[nxte->eval],
							d2ns(e->ntime), x_("0XX1")[e->eval]);
					nxte->type = CHECK_PNT;
					irsim_NewEdge(n, e);
					Deviate_Edge(n, e, n->npot, 0);
					n->npot = e->eval;
				}
				break;

			case CHECK_PNT:
				if (n->nflags & DEVIATED)
				{
					if (n->oldpot == e->eval)
						Merge_Edge(n, e, 0)
					else
						Deviate_SameVal(e)

					irsim_NewEdge(n, e);
					n->npot = e->eval;
				} else if (nxte->eval == e->eval)
				{
					register evptr  ne;

					if ((ne = n->events) != NULL)
						while(ne->nlink != NULL)
							ne = ne->nlink;

					if (ne == NULL || ne->ntime > nxte->ntime)
					{
						nxte->p.oldt = e->ntime;
						nxte->type = DELAY_EV;
						irsim_DelayEvent(e, d_t);
					} else
					{
						irsim_NewEdge(n, e);
						Deviate_Edge(n, e, n->npot, 0);
						n->npot = e->eval;
					}
				} else
				{
					irsim_NewEdge(n, e);
					Deviate_Edge(n, e, n->npot, 0);
					n->npot = e->eval;
				}
				break;

			case INP_EV:
			case XINP_EV:
				if (n->nflags & DEVIATED)
				{
					if (n->oldpot == e->eval)
						Merge_Edge(n, e, 0)
					else
						Deviate_SameVal(e);
				}
				else
					Deviate_Edge(n, e, n->npot, 0);

				n->npot = e->eval;
				irsim_NewEdge(n, e);
				break;

			default:
				ttyputmsg(x_("Unexpected Event 0x(%x)"), nxte->type);
		}
	}
}


#define CopyEvent(DST, SRC)		\
{					\
	(DST)->p.hist = (SRC)->p.hist;	\
	(DST)->ntime = (SRC)->ntime;	\
	(DST)->rtime = (SRC)->rtime;	\
	(DST)->delay = (SRC)->delay;	\
	(DST)->eval = (SRC)->eval;		\
}					\


static void UpdateCheck_Pnt(register evptr e)
{
	register nptr   n;
	register evptr  nxte;
	TimeType   d_t;
	short pot;

	n = e->enode;
	if ((nxte = n->events) != NULL)
	{
		while(nxte->nlink != NULL)
			nxte = nxte->nlink;
		d_t = nxte->ntime - e->ntime;
	}

	if (nxte == NULL || d_t > irsim_INC_RES)
	{
		irsim_DeleteNextEdge(n);
		if (n->npot == e->eval)
			Merge(n, e, CHK)
		else
			Deviate(n, e, e->eval, CHK);
	} else if (d_t == 0)
	{
		free_from_node(nxte, n);
		if (e->eval == nxte->eval)
		{
			n->curr = e->p.hist;
			if (DoingFault)  n->hchange = *n->curr, n->curr = &n->hchange;
			n->npot = e->eval;
			n->curr->t.r.delay = (short)nxte->delay;
			if (e->rtime == nxte->rtime)
				Merge_SameEdge(n, e, CHK)
			else
			{
				n->curr->t.r.rtime = nxte->rtime;
				Merge_Edge(n, e, CHK);
			}
			ReplacePunts(n);
		} else
		{
			irsim_DeleteNextEdge(n);
			pot = e->eval;		/* save old value temporarily */
			CopyEvent(e, nxte);	/* this is the new value! */
			irsim_NewEdge(n, e);
			n->npot = e->eval;
			Deviate_Edge(n, e, pot, CHK);
		}
	} else	/* 0 < d_t <= irsim_INC_RES */
	{
		if (n->nflags & DEVIATED)
		{
			irsim_DeleteNextEdge(n);
			if (n->npot == e->eval)
				Merge(n, e, CHK)
			else
				Deviate(n, e, e->eval, CHK);
		} else
		{
			register hptr   h;

			NEXTH(h, e->p.hist);
			if (nxte->eval == e->eval && (TimeType)h->htime > nxte->ntime)
			{
				e->type = DELAY_CHK;
				irsim_DelayEvent(e, d_t);
			} else
			{
				irsim_DeleteNextEdge(n);
				Deviate(n, e, e->eval, CHK);
			}
		}
	}
}


static void UpdateDelay_Chk(register evptr e)
{
	register nptr   n;
	register evptr  nxte;
	TimeType   d_t;
	short      pot;

	n = e->enode;
	if ((nxte = n->events) != NULL)
	{
		while(nxte->nlink != NULL)
			nxte = nxte->nlink;
		d_t = nxte->ntime - e->ntime;
	}

	if (nxte != NULL && d_t == 0 && nxte->eval == e->eval)
	{
		free_from_node(nxte, n);
		n->curr = e->p.hist;
		n->curr->t.r.delay = (short)nxte->delay;
		if (e->rtime == nxte->rtime)
			Merge_SameEdge(n, e, CHK)
		else
		{
			n->curr->t.r.rtime = nxte->rtime;
			Merge_Edge(n, e, CHK);
		}
		n->npot = e->eval;
		ReplacePunts(n);
	} else	/* event got punted (and maybe something else was queued) */
	{
		irsim_DeleteNextEdge(n);
		if (nxte != NULL && d_t == 0)
		{
			free_from_node(nxte, n);
			pot = e->eval;
			CopyEvent(e, nxte);
			Deviate_Edge(n, e, pot, CHK);
			irsim_NewEdge(n, e);
			n->npot = e->eval;
		}
		else					/* nothing else was queued */
			Deviate(n, e, e->eval, CHK);
	}
}


static void UpdateDelay_Ev(register evptr e)
{
	register nptr   n;
	register evptr  nxte;
	TimeType   d_t;
	short      pot;

	n = e->enode;
	if ((nxte = n->events) != NULL)
	{
		while(nxte->nlink != NULL)
			nxte = nxte->nlink;
		d_t = nxte->ntime - e->ntime;
	}

	if (nxte != NULL && d_t == 0 && nxte->eval == e->eval)
	{
		register hptr   h;

		free_from_node(nxte, n);
		NEXTH(h, n->curr);
		n->curr = h;
		Merge_SameEdge(n, e, CHK);
		n->npot = e->eval;
		ReplacePunts(n);
	} else		/* expected event was punted */
	{
		ttyputmsg(x_("Missed Glitch: %s => (%.2f %c) -> (~%.2f %c)"),
			pnode(n), d2ns(e->p.oldt), x_("0XX1")[e->eval],
				d2ns((e->ntime + e->p.oldt)/2), x_("0XX1")[n->npot]);

		irsim_DeleteNextEdge(n);
		if (nxte != NULL && d_t == 0)
		{
			free_from_node(nxte, n);
			irsim_NewEdge(n, nxte);
			pot = e->eval;
			CopyEvent(e, nxte);
			Deviate_Edge(n, e, d_t, CHK);
			n->npot = e->eval;
		}
		else
			Deviate(n, e, e->eval, CHK);
	}
}


static void update_nodes(evptr e)
{
	nptr   n;

	/* clear event lists */
	dev_evs = chk_evs = mrg_evs = mrg0_evs = inp_evs = xinp_evs =
	stim_evs = pend_evs = NULL;
	last_mrg = &mrg_evs;

	do
	{
#ifdef STATS
		{ extern int ev_hgm; if (ev_hgm) IncHistEvCnt((Uint) e->type); }
#endif
		switch(e->type)
		{
			case PUNTED:
				irsim_npunted_ev++;
				UpdateReval(e);
				break;

			case REVAL:
			case DECAY_EV:
				irsim_nreval_ev++;
				UpdateReval(e);
				break;

			case DELAY_CHK:
				irsim_ndelaychk_ev++;
				UpdateDelay_Chk(e);
				break;

			case DELAY_EV:
				irsim_ndelay_ev++;
				UpdateDelay_Ev(e);
				break;

			case CHECK_PNT:
				irsim_ncheckpt_ev++;
				UpdateCheck_Pnt(e);
				break;

			case XINP_EV:
				irsim_ncheckpt_ev++;
				n = e->enode;
				n->curr = e->p.hist;
				n->nflags &= ~(INPUT | ACTIVE_CL);
				e->nlink = xinp_evs;
				xinp_evs = e;
				ReplacePunts(n);
				break;

			case INP_EV:
				irsim_ncheckpt_ev++;
				n = e->enode;
				while(n->events && n->events->ntime != irsim_cur_delta)
					irsim_free_event(n->events);
				ReplacePunts(n);
				n->curr = e->p.hist;
				n->npot = e->eval;
				n->nflags |= INPUT;
				n->nflags &= ~DEVIATED;
				e->nlink = inp_evs;
				inp_evs = e;
				break;

			case STIM_INP:
				irsim_nstimuli_ev++;
				n = e->enode;
				n->npot = e->eval;
				n->curr = e->p.hist;
				n->nflags |= INPUT;
				e->nlink = stim_evs;
				stim_evs = e;
				break;

			case STIM_XINP:
				{
					register hptr  h;

					irsim_nstimuli_ev++;
					n = e->enode;
					n->curr = e->p.hist;
					n->nflags &= ~INPUT;
					NEXTH(h, n->curr);
					if (! irsim_EnqueueHist(n, h, STIMULI))
						n->nflags &= ~STIM;
					break;
				}

			case STIMULI:
				irsim_nstimuli_ev++;
				n = e->enode;
				n->npot = e->eval;
				n->curr = e->p.hist;
				e->nlink = stim_evs;
				stim_evs = e;
				break;

			case PENDING:
				e->nlink = pend_evs;
				pend_evs = e;
				break;
#ifdef notdef
			case PUNTED:
				irsim_npunted_ev++;
				n = e->enode;
				/* else: punt along with chk_pnt, the later handles it */
				if (n->c.event == NULL || n->c.event->ntime != e->ntime)
				{
					int  oval;

					free_from_node(e, n);
					irsim_NewEdge(n, e);
					oval = (n->nflags & DEVIATED) ? n->oldpot : n->npot;
					Deviate_Edge(n, e, oval, 0);
					n->npot = e->eval;
				}
				break;
#endif
			case CHNG_MODEL:
				curr_model = irsim_model_table[modelp->val];
				modelp = modelp->next;
				if (modelp != NULL)
					(void)irsim_EnqueueOther(CHNG_MODEL, (TimeType)modelp->htime);
				break;

#ifdef FAULT_SIM
			case TRIGGER_EV:
				stop_early = irsim_do_trigger(e);
				break;
#endif
			default:
				ttyputmsg(x_("update_nodes: bad event (%x) @ delta=%d for node %s"),
					e->type, irsim_cur_delta, pnode(e->enode));
		}
		irsim_npending -= 1;
	}
	while((e = e->flink) != NULL);

	*last_mrg = mrg0_evs;	/* join the 2 lists, order is important */
}


/*
 * Run through the various event lists and update the state of all transistors
 * controlled by the node that triggered the event.  Also mark all nodes that
 * need to be considered for evaluation as VISITED, this doesn't imply they
 * will be evaluated.
 */
static void UpdateTransistors(void)
{
	register evptr  e;
	register lptr   l;
	register tptr   t;

	for(e = dev_evs; e != NULL; e = e->nlink)
	{
		if ((e->type & O_DEV) == 0)		/* just deviated */
		{
			for(l = e->enode->ngate; l != NULL; l = l->next)
			{
				t = l->xtor;
				if (t->tflags & ACTIVE_T)
					t->state = compute_trans_state(t);
				t->source->nflags |= VISITED;	/* always mark src/drn */
				t->drain->nflags |= VISITED;
			}
		} else					/* was already deviated */
		{
			for(l = e->enode->ngate; l != NULL; l = l->next)
			{
				t = l->xtor;
				if (t->tflags & ACTIVE_T)
				{
					t->state = compute_trans_state(t);
					t->source->nflags |= VISITED;
					t->drain->nflags |= VISITED;
				}
			}
		}
	}

	for(e = mrg_evs; e != NULL; e = e->nlink)
	{
		for(l = e->enode->ngate; l != NULL; l = l->next)
		{
			t = l->xtor;
			if (t->tflags & ACTIVE_T)
			{
				t->state = compute_trans_state(t);
				t->source->nflags |= VISITED;
				t->drain->nflags |= VISITED;
			}
		}
	}

	for(e = chk_evs; e != NULL; e = e->nlink)
	{
		if ((e->type & (N_DEV | O_DEV)) == N_DEV)	/* just deviated */
		{
			for(l = e->enode->ngate; l != NULL; l = l->next)
			{
				t = l->xtor;
				if (t->tflags & ACTIVE_T)
					t->state = compute_trans_state(t);
				t->source->nflags |= VISITED;	/* always mark src/drn */
				t->drain->nflags |= VISITED;
			}
		}
	}

	for(e = inp_evs; e != NULL; e = e->nlink)
	{
		for(l = e->enode->ngate; l != NULL; l = l->next)
		{
			t = l->xtor;
			if (t->tflags & ACTIVE_T)
			{
				t->state = compute_trans_state(t);
				t->source->nflags |= VISITED;
				t->drain->nflags |= VISITED;
			}
		}
	}

	for(e = xinp_evs; e != NULL; e = e->nlink)
		e->enode->nflags |= VISITED;

	for(e = stim_evs; e != NULL; e = e->nlink)
	{
		for(l = e->enode->ngate; l != NULL; l = l->next)
		{
			t = l->xtor;
			if (t->tflags & ACTIVE_T)
			{
				t->state = compute_trans_state(t);
				t->source->nflags |= VISITED;
				t->drain->nflags |= VISITED;
			}
		}
	}

	irsim_VDD_node->nflags &= ~VISITED;
	irsim_GND_node->nflags &= ~VISITED;
}


/*
 * Build the connection list for node 'n'.  Determine (1) if all nodes in the
 * list are converged or whether this list is adjacent to a transistor whose
 * gate is deviated, (2) if there are any inactive nodes (or transistors) in
 * the connection list, and (3) whether any inputs (transistor gates) to the
 * connection list have a transition at the current time.  As a side effect
 * set the 'irsim_withdriven' flag.  The algorithm to construct the connection list
 * is the same as that used by irsim_BuildConnList.
 *
 * We must be careful not to report as an input transition nodes that stop
 * being inputs (i.e. hist->delay == 0 and hist->inp == 0).
 */
#define IsCurrTransition(H)			\
	((TimeType)(H)->htime == irsim_cur_delta && ((H)->inp == 1 || (H)->t.r.delay != 0))

static pstg GetConnList(register nptr n)
{
	register nptr  next, thisone, other, *inext;
	register lptr  l;
	register tptr  t;
	register hptr  h;
	register int   status;
	int            n_par = 0;
	static Stage   stage;

	stage.nd_list = n;
	stage.inp_list = NULL;

	status = ALL_MERGED;
	irsim_withdriven = FALSE;

	n->nflags &= ~VISITED;
	inext = &(stage.inp_list);

	if (NeedUpdate(n))
		(void)UpdateNode(n);
	if (IsDeviated(n))
		status &= ~ALL_MERGED;

	if (n->nflags & INPUT)
	{
		stage.flags = ALL_MERGED | ONLY_INPUT;
		return(&stage);
	}

	next = thisone = n->nlink = n;
	do
	{
		for(l = thisone->nterm; l != NULL; l = l->next)
		{
			t = l->xtor;
			if (! (t->tflags & ACTIVE_T))
			{
				if (t->ttype & GATELIST)
				{
					for(t = (tptr) t->gate; t != NULL; t = t->scache.t)
						if (NeedUpdate(t->gate))
							(void)UpdateNode(t->gate);
					t = l->xtor;
				} else if (NeedUpdate(t->gate))
					(void)UpdateNode(t->gate);

				t->state = compute_trans_state(t);
			} else if (status & ALL_MERGED)
			{
				if (t->ttype & GATELIST)
				{
					for(t = (tptr) t->gate; t != NULL; t = t->scache.t)
						if (t->gate->nflags & DEVIATED)
					{
						status &= ~ALL_MERGED;
						break;
					}
					t = l->xtor;
				} else if (t->gate->nflags & DEVIATED)
					status &= ~ALL_MERGED;
			}

			if (t->state == OFF)
				continue;

			if (t->tflags & CROSSED)
			{
				t->tflags &= ~CROSSED;
				continue;
			}

			t->scache.r = t->dcache.r = NULL;

			other = other_node(t, thisone);

			if (NeedUpdate(other))
				(void)UpdateNode(other);
			if (IsDeviated(other))
				status &= ~ALL_MERGED;

			if (other->nflags & INPUT)
			{
				irsim_withdriven = TRUE;
				if (other->nlink == NULL && 
					! (other->nflags & (ACTIVE_CL | POWER_RAIL)))
				{
					*inext = other;
					inext = &(other->nlink);
					*inext = other;		/* anything other than NULL */
				}
			} else
			{
				t->tflags |= CROSSED;
				if (other->nlink == NULL)
				{
					other->nflags &= ~VISITED;
					other->nlink = n;
					next->nlink = other;
					next = other;
					other->n.tran = t;
				}
				else if (irsim_model_num != LIN_MODEL)
					goto is_trans;
				else if (hash_terms(other->n.tran) == hash_terms(t))
				{
					register tptr  tran = other->n.tran;

					if (tran->tflags & PARALLEL)
						t->dcache.t = par_list(tran);
					else
					{
						if (n_par >= MAX_PARALLEL)
						{
							irsim_WarnTooManyParallel(0, 0);
							t->tflags |= PBROKEN;
							continue;
						}
						tran->n_par = n_par++;
						tran->tflags |= PARALLEL;
					}
					par_list(tran) = t;
					t->tflags |= PBROKEN;
				} else
				{
					t->tflags |= BROKEN;
					continue;
				}
			}
is_trans:
			if (t->ttype & GATELIST)
			{
				for(t = (tptr) t->gate; t; t = t->scache.t)
				{
					h = t->gate->curr;
					if (IsCurrTransition(h))
					{
						t = l->xtor;
						status |= INP_TRANS;
						break;
					}
				}
				t = l->xtor;
			} else
			{
				h = t->gate->curr;
				if (IsCurrTransition(h))
					status |= INP_TRANS;
			}
		}		/* end of for each nterm */
	}
	while((thisone = thisone->nlink) != n);

	next->nlink = NULL;
	*inext = NULL;
	stage.flags = status;

	return(&stage);
}


/*
 * Activate the gate of a transistor to act as stimuli.  First check
 * that the node in question is not part of an active cluster or is
 * already acting as stimuli.
 */
#define	StimulateGate(GATE)				\
{							\
	if (NeedUpdate(GATE))				\
	{							\
		register hptr nextH = UpdateNode(GATE);	\
		if (irsim_EnqueueHist(GATE, nextH, STIMULI))	\
			(GATE)->nflags |= STIM;			\
	}							\
}


/*
 * Activate all nodes and transistors in the connection list of node 'nd',
 * scheduling pending events on the outputs and activating any inputs as
 * stimuli.  This routine assumes that the connection list has already been
 * built GetConnList.  Also activate all INPUT nodes connected to the stage.
 * The routine returns TRUE if any of the nodes in the connection list has
 * any pending punted events, else FALSE;
 */
static void ActivateStage(pstg stg)
{
	register nptr  n;

	n = stg->nd_list;
	do
	{
		register lptr  l;
		register tptr  t;

		if (! (n->nflags & ACTIVE_CL))
			ActivateNode(n);

		for(l = n->nterm; l != NULL; l = l->next)
		{
			t = l->xtor;
			if (t->tflags & ACTIVE_T)
				continue;
			t->tflags |= ACTIVE_T;
			if (t->ttype & GATELIST)
			{
				for(t = (tptr) t->gate; t != NULL; t = t->scache.t)
					StimulateGate(t->gate);
			}
			else
				StimulateGate(t->gate);
		}
	}
	while((n = n->nlink) != NULL);

	n = stg->inp_list;		/* traverse inputs connected to stage */
	while(n != NULL)
	{
		register nptr  next;

		if (! (n->nflags & (ACTIVE_CL | POWER_RAIL)))
			(void)ActivateNode(n);

		next = n->nlink;
		n->nlink = NULL;
		n = next;
	}
}


/*
 * Merge any 'future' punted events from previous runs with the current
 * history for the node.  Note that events created prior to irsim_cur_delta are
 * simply discarded (if they belong in the history the should have been
 * recreated and re-punted).
 */
static void MergePunts(nptr nd)
{
	register hptr  h, p, nx;

	/* skip past any punted events already in the history */
	for(nx = nd->curr; nx->next->punt; nx = nx->next);

	p = nd->t.punts;
	do
	{
		h = p;
		p = p->next;
		if (QPTIME(h) < irsim_cur_delta)		/* in the past, free it */
		{
			h->next = irsim_freeHist;
			irsim_freeHist = h;
		} else					/* in the future, merge */
		{
			h->next = nx->next;
			nx->next = h;
			nx = h;
		}
	}
	while(p != NULL);

	nd->t.punts = NULL;
}


/*
 * Deactivate all nodes on the connection list and destroy the list.  If a
 * node has any gate connections then turn it into a stimulus node.  Also
 * deactivate any non-OFF transistor on the connection list and move any
 * punted events from the previous simulation that have yet to be enqueued
 * back into the history.  We can't deactivate OFF transistors because some
 * of them may be boundary transistors.
 * For all transistors in the stage, clear the src/drn caches and any flags
 * that may have been set when building stage.
 */
static void DeactivateStage(pstg stg, nptr skipnd)
{
	register nptr  n, next;

	for(n = stg->nd_list; n != NULL; n = next)
	{
		register lptr  l;
		register tptr  t;

		next = n->nlink;
		n->nlink = NULL;

#ifdef INCDEBUG
		if (n->nflags & WATCHED)
			ttyputmsg(x_("%.2f: Deactivate %s"), d2ns(irsim_cur_delta),pnode(n));
#endif
		if (n->nflags & ACTIVE_CL)
		{
			if (n->c.event != NULL && n->c.event->ntime > irsim_cur_delta)
				irsim_DequeueEvent(n);
			n->nflags &= ~ACTIVE_CL;
			if (n->t.punts != NULL)
				MergePunts(n);
		}
		if (n->ngate != NULL && ! (n->nflags & STIM) && n != skipnd)
		{
			hptr  h;
			NEXTH(h, n->curr);
			if (irsim_EnqueueHist(n, h, STIMULI))
				n->nflags |= STIM;
		}

		for(l = n->nterm; l != NULL; l = l->next)
		{
			t = l->xtor;
			t->dcache.r = NULL;		/* may be set for parallel xtors */

			if ((t->tflags & ACTIVE_T) && (t->state != OFF ||
				(other_node(t, n)->nflags & INPUT)) )
					t->tflags = 0;
			else
				t->tflags &= ~(BROKEN | PBROKEN | PARALLEL);
		}
	}

	for(n = stg->inp_list; n != NULL; n = next)
	{
		next = n->nlink;
		n->nlink = NULL;
	}
}



static void UndoStage(pstg stg)
{
	register nptr  n, next;
	register lptr  l;
	register tptr  t;

	for(n = stg->nd_list; n != NULL; n = next)
	{
		next = n->nlink;
		n->nlink = NULL;

		for(l = n->nterm; l != NULL; l = l->next)
		{
			t = l->xtor;
			t->dcache.r = NULL;
			t->tflags &= ~(BROKEN | PBROKEN | PARALLEL);
		}
	}

	for(n = stg->inp_list; n != NULL; n = next)
	{
		next = n->nlink;
		n->nlink = NULL;
	}
}


/*
 * Evaluate the source/drain of all active transistors controlled by 'nd'.
 * Activate/Deactivate clusters as required.  Return TRUE if there are any
 * active transistors controlled by this node, FALSE otherwise.
 * If force is TRUE then evaluate even if all nodes in the clusters are
 * merged.
 */
static int EvalSrcDrn(nptr nd, int force)
{
	register lptr  l;
	register int   found;
	register nptr  nterm;
	pstg           stg;

	found = FALSE;
	irsim_cur_node = nd;
	for(l = nd->ngate; l != NULL; l = l->next)
	{
		tptr  t = l->xtor;

		if (!(t->tflags & ACTIVE_T))
			continue;

		found = TRUE;
		nterm = t->source;
		if (nterm->nflags & VISITED)
		{
			stg = GetConnList(nterm);
			if (stg->flags & ONLY_INPUT)
			{
				if (! (nterm->nflags & (ACTIVE_CL | POWER_RAIL)) && 
					nd->nflags & DEVIATED)
				(void)ActivateNode(nterm);
			}
			else if ((stg->flags & ALL_MERGED) && ! force)
				DeactivateStage(stg, (nptr) NULL);
			else
			{
				ActivateStage(stg);
				(*curr_model)(nterm);
			}
		}

		nterm = t->drain;
		if (nterm->nflags & VISITED)
		{
			stg = GetConnList(nterm);
			if (stg->flags & ONLY_INPUT)
			{
				if (! (nterm->nflags & (ACTIVE_CL | POWER_RAIL)) && 
					nd->nflags & DEVIATED)
						ActivateNode(nterm);
			}
			else if ((stg->flags & ALL_MERGED) && ! force)
				DeactivateStage(stg, (nptr) NULL);
			else
			{
				ActivateStage(stg);
				(*curr_model)(nterm);
			}
		}
	}
	return(found);
}


/*
 * Evaluate the source/drain of ALL transistors controlled by 'nd', which just
 * deviated.  Different things should be done if a transition caused the node
 * to deviate or a lack of transition.
 */
static void EvalJustDeviated(nptr nd, int has_trans)
{
	pstg  stg;
	lptr  l;

	irsim_cur_node = nd;
	for(l = nd->ngate; l != NULL; l = l->next)
	{
		register nptr  nterm;
		register tptr  t;

		t = l->xtor;
		nterm = t->source;
		if (nterm->nflags & VISITED)
		{
			stg = GetConnList(nterm);
			if (stg->flags & ONLY_INPUT)
			{
				if (! (nterm->nflags & (ACTIVE_CL | POWER_RAIL)))
					ActivateNode(nterm);
			} else
			{
				ActivateStage(stg);
				if (has_trans || (stg->flags & INP_TRANS))
					(*curr_model)(nterm);
				else
					UndoStage(stg);
			}
		}

		nterm = t->drain;
		if (nterm->nflags & VISITED)
		{
			stg = GetConnList(nterm);
			if (stg->flags & ONLY_INPUT)
			{
				if (! (nterm->nflags & (ACTIVE_CL | POWER_RAIL)))
					ActivateNode(nterm);
			} else
			{
				ActivateStage(stg);
				if (has_trans || (stg->flags & INP_TRANS))
					(*curr_model)(nterm);
				else
					UndoStage(stg);
			}
		}

		/* may happen if src/drn are inputs now */
		if (! (t->tflags & ACTIVE_T))
		{
			t->tflags |= ACTIVE_T;
			if (t->ttype & GATELIST)
			{
				for(t = (tptr) t->gate; t != NULL; t = t->scache.t)
					StimulateGate(t->gate);
				t = l->xtor;
			}
			t->state = compute_trans_state(t);
		}
	}
}


static void EvalInputs(void)
{
	register evptr  ev;
	register nptr   n;

	for(ev = inp_evs; ev != NULL; ev = ev->nlink)
	{
		register nptr   other;
		register lptr   l;
		register tptr   t;
		int             act_tran = FALSE;

		n = irsim_cur_node = ev->enode;
		n->nflags &= ~ACTIVE_CL;	/* assume it won't be active */

		for(l = n->nterm; l != NULL; l = l->next)
		{
			pstg  stg;

			t = l->xtor;
			other = other_node(t, n);
			stg = GetConnList(other);
			if (stg->flags & ALL_MERGED)
			{
				if (! (stg->flags & ONLY_INPUT))
					DeactivateStage(stg, (nptr) NULL);
				else if (t->tflags & ACTIVE_T)
					act_tran = TRUE;
			} else
			{
				ActivateStage(stg);
				(*curr_model)(other);
			}
		}
		if (act_tran == TRUE && ! (n->nflags & (POWER_RAIL | ACTIVE_CL)))
		{
			register hptr  h;

			NEXTH(h, n->curr);
			if (irsim_EnqueueHist(n, h, CHECK_PNT))
				n->nflags |= ACTIVE_CL;
		}
	}

	for(ev = inp_evs; ev != NULL; ev = ev->nlink)
	{
		int   n_active;

		n = irsim_cur_node = ev->enode;
		n_active = EvalSrcDrn(n, FALSE);

		/* 
		* if node is not active anymore and it controls any 
		* active transistors turn it into a stimulus node
		*/
		if (n_active == TRUE && ! (n->nflags & ACTIVE_CL))
		{
			register hptr  h;

			NEXTH(h, n->curr);
			if (irsim_EnqueueHist(n, h, STIMULI))
				n->nflags |= STIM;
		}
	}
}


static void EvalStimulus(void)
{
	evptr  ev;
	int    n_active;
	nptr   nd;

	for(ev = stim_evs; ev != NULL; ev = ev->nlink)
	{
		nd = ev->enode;
		n_active = EvalSrcDrn(nd, FALSE);

		if (! n_active)			/* no active transistors */
			nd->nflags &= ~STIM;
		else if (nd->nflags & STIM)		/* enqueue next transition */
		{
			register hptr  h;

			NEXTH(h, nd->curr);
			if (! irsim_EnqueueHist(nd, h, STIMULI))
				nd->nflags &= ~STIM;
		}
	}
}


static void EvalXinputs(void)
{
	register evptr  ev;
	register nptr   n;
	pstg            stg;

	for(ev = xinp_evs; ev != NULL; ev = ev->nlink)
	{
		n = irsim_cur_node = ev->enode;
		if (n->nflags & VISITED)
		{
			stg = GetConnList(n);
			if (stg->flags & ALL_MERGED)
				DeactivateStage(stg, (nptr) NULL);
			else
			{
				ActivateStage(stg);
				(*curr_model)(n);
			}
		}
	}
}


static void EvalEventList(void)
{
	register evptr  e;
	register nptr   n;
	int             status;
	pstg            stg;

	for(e = dev_evs; e != NULL; e = e->nlink)
	{
		n = e->enode;
		if ((e->type & O_DEV) == 0)
			EvalJustDeviated(n, TRUE);
		else
			(void)EvalSrcDrn(n, FALSE);

		/* if the node is still active schedule its next CHECK_PNT event */
		if ((n->nflags & ACTIVE_CL) && (e->type & CHK))
		{
			hptr  h;

			NEXTH(h, n->curr);
			(void)irsim_EnqueueHist(n, h, CHECK_PNT);
		}
	}

	for(e = mrg_evs; e != NULL; e = e->nlink)
	{
		status = ((e->type & (SAME_T | O_DEV)) == O_DEV) ? TRUE : FALSE;

		if (EvalSrcDrn(e->enode, status))
			e->type |= N_ACTV;
	}

	for(e = mrg_evs; e != NULL; e = e->nlink)
	{
		n = e->enode;
		stg = GetConnList(n);
		if (stg->flags & ALL_MERGED)
			DeactivateStage(stg, (e->type & N_ACTV) ? (nptr) NULL : n);
		else
			UndoStage(stg);

		if ((n->nflags & ACTIVE_CL) && (e->type & CHK))
		{
			hptr  h;

			NEXTH(h, n->curr);
			(void)irsim_EnqueueHist(n, h, CHECK_PNT);
		}
	}

	for(e = chk_evs; e != NULL; e = e->nlink)
	{
		n = e->enode;
		switch(e->type & (EDGE| O_DEV | N_DEV))
		{
			case 0:
				ttyputmsg(x_("warning: case 0 time=%.2f for %s"),
					d2ns(irsim_cur_delta), pnode(n));
				break;

			case O_DEV:	/* Node just merged but no transition */
				stg = GetConnList(n);
				if (stg->flags & ALL_MERGED)
					DeactivateStage(stg, (nptr) NULL);
				else
					UndoStage(stg);
				break;

			case N_DEV:	/* Node just deviated but no transition */
				n = e->enode;
				EvalJustDeviated(n, FALSE);
				break;

			case O_DEV | N_DEV:/* Node stays deviated but no transition */
				break;		/* -- do nothing -- */

			default:
				ttyputmsg(x_("bad chk event (0x%2x) @ t=%.2f"), e->type,
					d2ns(irsim_cur_delta));
		}

		if ((n->nflags & ACTIVE_CL) && (e->type & CHK))
		{
			hptr  h;

			NEXTH(h, n->curr);
			(void)irsim_EnqueueHist(n, h, CHECK_PNT);
		}
	}
}



/*
 * Events pending from a previous simulation run.  If the node is active at
 * this point, simply drop the event (the node is being resimulated) otherwise
 * re-schedule the pending event (as if it had been simulated).
*/
static void EvalPending(void)
{
	register evptr  ev;
	register nptr   n;

	for(ev = pend_evs; ev != NULL; ev = ev->nlink)
	{
		n = ev->enode;
		if (! (n->nflags & ACTIVE_CL))
			irsim_enqueue_event(n, (int) ev->eval, (long) ev->delay, (long) ev->rtime);
	}
}


static void incstep(TimeType stop_time)
{
	evptr  evlist;
	TimeType   step_t, refresh;

	refresh = (stop_time < 10) ? 1 : stop_time / 10;
	step_t = irsim_cur_delta + refresh;
	if (DoingFault) step_t = irsim_max_time;		/* never refresh */

	while((evlist = irsim_get_next_event(stop_time)) != NULL)
	{
		update_nodes(evlist);
		UpdateTransistors();

		EvalEventList();

		if (inp_evs)	    EvalInputs();
		if (xinp_evs)	    EvalXinputs();
		if (stim_evs)	    EvalStimulus();
		if (pend_evs)	    EvalPending();

		FreeEventList(evlist);

		if (irsim_cur_delta >= step_t)
		{
			do
			{
				ttyputmsg(x_("time = %d.0"), (int) d2ns(step_t));
				(void)fflush(stdout);
				step_t += refresh;
			}
			while(irsim_cur_delta >= step_t);
			if (irsim_analyzerON)
				irsim_UpdateWindow(irsim_cur_delta - 1);
		}
#ifdef FAULT_SIM
		if (stop_early || stopping(STOPREASONSIMULATE))
			return;
#endif
	}
	irsim_cur_delta = stop_time;
#ifndef FAULT_SIM
	if (irsim_analyzerON)
		irsim_UpdateWindow(irsim_cur_delta);
#endif
  }


static int fix_inc_nodes(nptr nd)
{
	register hptr  h;
	register lptr  l;

	if (nd->nflags & (ALIAS | MERGED))
		return(0);

	if ((nd->nflags & ACTIVE_CL) && nd->t.punts != NULL)  /* shouldn't be */
	{
		for(h = nd->t.punts; h->next != NULL; h = h->next);
		h->next = irsim_freeHist;
		irsim_freeHist = nd->t.punts;
	}

	if (nd->nflags & (WAS_ACTIVE | CHANGED))
		nd->t.cause = irsim_inc_cause;

	nd->nflags &= ~(VISITED|CHANGED|DEVIATED|STIM|ACTIVE_CL|WAS_ACTIVE);

	NEXTH(h, nd->curr);
	if (h != irsim_last_hist)		/* inconsistent ? */
	{
		register hptr  p;

		do
		{
			p = h;
			NEXTH(h, h);
		} while(h != irsim_last_hist);
		nd->curr = h = p;
	}
	else
	h = nd->curr;

	nd->c.time = (TimeType)h->htime;
	nd->npot = h->val;
	if (h->inp)
		nd->nflags |= INPUT;
	else
		nd->nflags &= ~INPUT;

	for(l = nd->ngate; l != NULL; l = l->next)	/* fix transistors */
	{
		register tptr  t = l->xtor;
		t->state = compute_trans_state(t);
		t->tflags &= ~ACTIVE_T;
	}
	for(l = irsim_on_trans; l != NULL; l = l->next)
		l->xtor->tflags &= ~ACTIVE_T;

	return(0);
}


/*
 * Start evaluating/activating the changed nodes.  This will schedulle other
 * events to get things rolling.  Special care is needed for changed nodes
 * that became inputs at time 0; we must ensure that their adjacent (src/drn)
 * transistors have the proper state, so things work out when the node
 * becomes undriven (if at all) and revaluated properly,
 */
static void startup_isim(nptr n)
{
	pstg  stg;

	stg = GetConnList(n);
	ActivateStage(stg);
	if (stg->flags & INP_TRANS)	/* transition at time 0 => evaluate */
		(*curr_model)(n);
	else if (stg->flags & ONLY_INPUT)	/* node is an input at time 0 */
	{
		register lptr  l;
		register tptr  t;

		for(l = n->nterm; l != NULL; l = l->next)
		{
			t = l->xtor;
			t->state = compute_trans_state(t);
		}
		UndoStage(stg);
	}
	else				/* no transitions at time 0, undo */
		UndoStage(stg);
}


/*
 * Main incremental simulation routine.  Move back to time 0, changing all
 * pending events to PENDING type.  Activate the clusters of all the
 * nodes in 'changed nodes' list and start simulating until we reach the
 * current time.  Make sure that the whole network is in a consistent state
 * before returning.
 */
void irsim_incsim(iptr ch_list)
{
	TimeType  stop_time;

	o_nevals = irsim_nevals;
	irsim_nevals = irsim_i_nevals;
	o_nevent = irsim_nevent;
	irsim_sm_stat |= INCR_SIM;

	stop_time = irsim_cur_delta;
	irsim_cur_delta = irsim_sim_time0;

	(void)irsim_back_sim_time(irsim_cur_delta, TRUE);

	modelp = irsim_first_model;			/* initialize the model */
	curr_model = irsim_model_table[modelp->val];
	modelp = modelp->next;
	if (modelp != NULL)
		(void)irsim_EnqueueOther(CHNG_MODEL, (TimeType)modelp->htime);

	if (ch_list != NULL)
	{
		register iptr  ip;

		if (stop_time != 0)		/* work to do ? */
		{
			for(ip = ch_list; ip != NULL; ip = ip->next)
				ip->inode->nflags |= VISITED;	/* set VISITED on all nodes */
		} else
		{
			while((ip = ch_list) != NULL)
			{
				ch_list = ch_list->next;
				ip->inode->nflags &= ~(VISITED);
				FreeInput(ip);
			}
		}

		while((ip = ch_list) != NULL)
		{
			ch_list = ch_list->next;
			if (ip->inode->nflags & VISITED)
				startup_isim(ip->inode);
			FreeInput(ip);
		}
	}

	incstep(stop_time);

	irsim_rm_inc_events(FALSE);

	irsim_walk_net(fix_inc_nodes);

	irsim_sm_stat &= ~INCR_SIM;
	irsim_i_nevals = irsim_nevals;
	irsim_nevals = o_nevals;
	irsim_nevent = o_nevent;
}


#ifdef FAULT_SIM
void irsim_init_faultsim(void)
{
	o_nevals = irsim_nevals;
	irsim_nevals = irsim_i_nevals;
	o_nevent = irsim_nevent;
	irsim_sm_stat |= INCR_SIM;
	old_cur_delta = irsim_cur_delta;
	irsim_cur_delta = irsim_sim_time0;
	pending_evs = irsim_back_sim_time(irsim_cur_delta, TRUE + TRUE);
	irsim_fault_mode = TRUE;
}


void irsim_end_faultsim(void)
{
	irsim_walk_net(fix_inc_nodes);
	irsim_cur_delta = old_cur_delta;
	irsim_requeue_events(pending_evs, TRUE);
	irsim_fault_mode = FALSE;
	irsim_sm_stat &= ~INCR_SIM;
	irsim_i_nevals = irsim_nevals;
	irsim_nevals = o_nevals;
	irsim_nevent = o_nevent;
}


static int fix_fault_nodes(nptr nd)
{
	register lptr  l;

	if (! (nd->nflags & (ALIAS | MERGED)))
	{
		nd->nflags &= ~(VISITED|CHANGED|DEVIATED|STIM|ACTIVE_CL|WAS_ACTIVE);

		if (nd->curr == &nd->hchange)
		{
			register hptr  h;

			NEXTH(h, nd->curr);
			nd->curr = (h == irsim_last_hist) ? &nd->head : h;
		}

		/* only mark transistors as non-activate for next faultsim */
		for(l = nd->ngate; l != NULL; l = l->next)
		{
			register tptr  t = l->xtor;
			t->tflags &= ~ACTIVE_T;
		}
		for(l = irsim_on_trans; l != NULL; l = l->next)
			l->xtor->tflags &= ~ACTIVE_T;
	}
	return(0);
}


void irsim_faultsim(nptr n)
{
	TimeType   stop_time;

	stop_time = old_cur_delta;

	irsim_cur_delta = irsim_sim_time0;
	stop_early = 0;

	modelp = irsim_first_model;			/* initialize the model */
	curr_model = irsim_model_table[modelp->val];
	modelp = modelp->next;
	if (modelp != NULL)
		(void)irsim_EnqueueOther(CHNG_MODEL, (TimeType) modelp->htime);

	irsim_setup_triggers();

	n->nflags |= (VISITED | CHANGED);
	startup_isim(n);

	incstep(stop_time);

	irsim_rm_inc_events(TRUE);

	irsim_walk_net(fix_fault_nodes);
}
#endif
